from typing import List, Any

from graphql import execute, parse

from graphql_utilities.decorators import run_only_once
from tests.helpers import assert_no_errors
from tests.schema import schema


def test_run_only_once():
    runs: List[Any] = []

    class Middleware:
        @run_only_once
        def resolve(self, next_, root, info, *args, **kwargs):
            runs.append(True)
            return next_(root, info, *args, **kwargs)

    query = '{ field_1_str field_2_int field_3_obj { field_3_obj_sub_1 { xxx } } }'

    result = execute(schema=schema, document=parse(query), middleware=[Middleware()])
    assert_no_errors(result)
    assert len(runs) == 1, "middleware should run only once"



def run_only_once(resolve_func):
    """
    Make sure middleware is run only once,
    this is done by setting a flag in the `context` of `ResolverInfo`

    Example:
        class AuthenticationMiddleware:
            @run_only_once
            def resolve(self, next, root, info, *args, **kwargs):
                pass
    """
    from functools import wraps

    # Use closure to keep track of middleware
    middleware_run = dict()

    @wraps(resolve_func)
    def wrapper(self, next, root, info, *args, **kwargs):
        middleware_name = self.__class__.__name__
        if not middleware_run.get(middleware_name):
            middleware_run[middleware_name] = True
            return resolve_func(self, next, root, info, *args, **kwargs)

        return next(root, info, *args, **kwargs)

    return wrapper



from graphql import GraphQLSchema, GraphQLObjectType, GraphQLField, GraphQLString, GraphQLInt

schema = GraphQLSchema(
    query=GraphQLObjectType(
        name='RootQueryType',
        fields={
            'field_1_str': GraphQLField(
                GraphQLString,
                resolve=lambda obj, info: 'str_1_field'),
            'field_2_int': GraphQLField(
                GraphQLInt,
                resolve=lambda obj, info: 2),
            'field_3_obj': GraphQLField(
                GraphQLObjectType(
                    name='Field3Object',
                    fields={
                        'field_3_obj_sub_1': GraphQLField(GraphQLObjectType(
                            name='Field3ObjectSubfield1',
                            fields={
                                'xxx': GraphQLField(GraphQLString)
                            }
                        ))
                    }
                ),
                resolve=lambda obj, info: {'field_3_obj': {'field_3_obj_sub_1': '1_sub_obj_3_field'}}
            ),
            'cat': GraphQLField(
                GraphQLObjectType(
                    name='CatObject',
                    fields={
                        'meow': GraphQLField(GraphQLString)
                    },
                ),
                resolve=lambda obj, info: {"meow": "woem"}
            )
        }))



from graphql import Visitor, GraphQLError, BREAK, IDLE, ValidationContext
from graphql.language import ast


class DepthAnalysisVisitor(Visitor):
    def __init__(self, context: ValidationContext, max_depth: int = 10):
        self.max_depth = max_depth
        self.context = context

    def enter(self, node, key, parent, path, *args):
        if isinstance(node, ast.NameNode):
            depth = path.count("selection_set")
            if depth > self.max_depth:
                self.context.report_error(GraphQLError(f"Reached max depth of {self.max_depth}"))
                return BREAK

        return IDLE



from typing import Any, Dict, Union, List, Optional

from graphql import GraphQLError, ExecutionContext, GraphQLSchema, DocumentNode, \
    GraphQLFieldResolver, GraphQLTypeResolver, Middleware, ValidationContext, TypeInfo, visit

from graphql_utilities.visitor import DepthAnalysisVisitor

ContextValue = Optional[Dict[str, Any]]


class ValidationAbortedError(RuntimeError):
    """
    Taken from: graphql-core-next/graphql.validation.validate
    Error when a validation has been aborted (error limit reached).
    """
    pass


class ExtendedExecutionContext(ExecutionContext):
    @classmethod
    def build(
            cls,
            schema: GraphQLSchema,
            document: DocumentNode,
            root_value: Any = None,
            context_value: Any = None,
            raw_variable_values: Dict[str, Any] = None,
            operation_name: str = None,
            field_resolver: GraphQLFieldResolver = None,
            type_resolver: GraphQLTypeResolver = None,
            middleware: Middleware = None,
    ) -> Union[List[GraphQLError], ExecutionContext]:
        # Build execution context after visiting it
        errors: List[GraphQLError] = []

        def on_error(error: GraphQLError) -> None:
            errors.append(error)
            raise ValidationAbortedError

        depth_analysis = context_value.get("depth_analysis")
        if depth_analysis:
            max_depth = depth_analysis.get("max_depth", 10)
            context = ValidationContext(schema=schema, ast=document, type_info=TypeInfo(schema), on_error=on_error)

            try:
                visit(document, DepthAnalysisVisitor(context=context, max_depth=max_depth))
            except ValidationAbortedError:
                pass

        if errors:
            return errors

        return super(ExtendedExecutionContext, cls).build(
            schema,
            document,
            root_value,
            context_value,
            raw_variable_values,
            operation_name,
            field_resolver,
            type_resolver,
            middleware
        )



import os

import setuptools


def get_version():
    package_init = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'graphql_utilities', '__init__.py')
    with open(package_init) as f:
        for line in f:
            if line.startswith('__version__ ='):
                return line.split('=')[1].strip().strip('"\'')


def get_long_description():
    with open("README.md", "r") as fh:
        return fh.read()


setuptools.setup(
    name="graphql-utilities",
    version=get_version(),
    author="Melvin Koh",
    author_email="melvinkcx@gmail.com",
    description="Collection of utilities, middleware, decorators for graphql-core>=3.0",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/melvinkcx/graphql-utilities",
    packages=setuptools.find_packages(exclude=["tests.*"]),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.5',
        'Topic :: Software Development :: Libraries'
    ],
    install_requires=['graphql-core>=3.0'],
    python_requires=">=3.6,<4",
    extras_require={
        'dev': [
            'pytest>=5.3.4,<6',
            'flake8>=3.7.9,<4',
            'pytest-describe>=0.12,<1'
        ]
    }
)



from graphql import execute, parse

from graphql_utilities.execution import ExtendedExecutionContext
from tests.helpers import assert_no_errors, assert_has_errors, assert_has_exactly_n_errors, assert_has_error_message
from tests.schema import schema


def describe_depth_analysis():
    def _execute_query(max_depth: int):
        query = '{ field_1_str field_2_int field_3_obj { field_3_obj_sub_1 { xxx } } }'
        return execute(schema=schema, document=parse(query),
                       context_value={"depth_analysis": {
                           "max_depth": max_depth
                       }},
                       execution_context_class=ExtendedExecutionContext)

    def test_query_less_than_max_depth_allowed():
        result = _execute_query(10)
        assert_no_errors(result)

    def test_query_exceeding_max_depth():
        result = _execute_query(2)
        assert_has_errors(result)
        assert_has_exactly_n_errors(result, n=1)
        assert_has_error_message(result, error_message="Reached max depth of 2")



from graphql import ExecutionResult


def assert_no_errors(result: ExecutionResult) -> None:
    assert not result.errors, "Results should contain errors"


def assert_no_data(result: ExecutionResult) -> None:
    assert not result.data, "Results shouldn't contain data"


def assert_has_data(result: ExecutionResult) -> None:
    assert not result.data, "Results should contain data"


def assert_has_errors(result: ExecutionResult) -> None:
    assert result.errors, "ExecutionResult should contain errors"


def assert_has_exactly_n_errors(result: ExecutionResult, n: int) -> None:
    assert len(result.errors) == n, f"ExecutionResult should contain exactly {n} error(s)"


def assert_has_error_message(result: ExecutionResult, error_message: str) -> None:
    has_error_message = False
    for err in result.errors:
        if err.message.find(error_message) != -1:
            has_error_message = True

    assert has_error_message, f"'{error_message}' not found"
